package nullfree.linkedlist.std;

import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import nullfree.linkedlist.HeadAndTail;
import nullfree.linkedlist.Node;
import nullfree.linkedlist.NodeFactory;
import nullfree.linkedlist.polymorph.PolymorphNodeFactory;

public class LinkedList<T> implements List<T> {
	private HeadAndTail<T> hat;
	private Node<T> find(int ix) {
		Node<T> res = hat.getHead();
		for(int i = 0; i < ix; i++) {
			res = res.getNext();
		}
		return res;
	}
	public LinkedList(T first) {
        NodeFactory<T> nf = new PolymorphNodeFactory<T>();
        hat = nf.createFirstNode(first);
	}
	public boolean add(T o) {
		hat.getTail().insertValue(o, hat);
		return true;
	}
	public void add(int ix, T o) {
		find(ix).insertValue(o, hat);
	}
	public boolean addAll(Collection<? extends T> col) {
		for(T o : col) {
			hat.getTail().insertValue(o, hat);
		}
		return true;
	}
	public boolean addAll(int ix, Collection<? extends T> col) {
		Node<T> tmp = find(ix);
		for(T o : col) {
			tmp = tmp.insertValue(o, hat);
		}
		return true;
	}
	public void clear() {
		throw new UnsupportedOperationException("Can not clear a null free linked list");
	}
	public boolean contains(Object o) {
		Node<T> tmp = hat.getHead();
		if(tmp.getValue().equals(o)) return true;
		while(tmp.hasNext()) {
			tmp = tmp.getNext();
			if(tmp.getValue().equals(o)) return true;
		}
		return false;
	}
	public boolean containsAll(Collection<?> col) {
		for(Object o : col) {
			if(!contains(o)) return false;
		}
		return true;
	}
	public T get(int ix) {
		return find(ix).getValue();
	}
	public int indexOf(Object o) {
		int ix = 0;
		Node<T> tmp = hat.getHead();
		if(tmp.getValue().equals(o)) return ix;
		while(tmp.hasNext()) {
			tmp = tmp.getNext();
			ix++;
			if(tmp.getValue().equals(o)) return ix;
		}
		return -1;
	}
	public boolean isEmpty() {
		return false;
	}
	private class LinkedListIterator implements Iterator<T>, ListIterator<T> {
		private Node<T> curr;
		private boolean eol;
		private boolean bol;
		public LinkedListIterator(Node<T> start) {
			curr = start;
			eol = false;
			bol = false;
		}
		public boolean hasNext() {
			return !eol;
		}
		public T next() {
			T res = curr.getValue();
			eol = !curr.hasNext();
			curr = curr.getNext();
			return res;
		}
		public void remove() {
			curr.remove(hat);
		}
		public void add(T o) {
			curr.insertValue(o, hat);
		}
		public boolean hasPrevious() {
			return !bol;
		}
		public int nextIndex() {
			return indexOf(curr) + 1;
		}
		public T previous() {
			T res = curr.getValue();
			bol = !curr.hasPrev();
			curr = curr.getPrev();
			return res;
		}
		public int previousIndex() {
			return indexOf(curr) - 1;
		}
		public void set(T o) {
			curr.setValue(o);
		}
	}
	public Iterator<T> iterator() {
		return new LinkedListIterator(hat.getHead());
	}
	public int lastIndexOf(Object o) {
		int res = -1;
		int ix = 0;
		Node<T> tmp = hat.getHead();
		if(tmp.getValue().equals(o)) res = ix;
		while(tmp.hasNext()) {
			tmp = tmp.getNext();
			ix++;
			if(tmp.getValue().equals(o)) res = ix;
		}
		return res;
	}
	public ListIterator<T> listIterator() {
		return new LinkedListIterator(hat.getHead());
	}
	public ListIterator<T> listIterator(int ix) {
		return new LinkedListIterator(find(ix));
	}
	public boolean remove(Object o) {
		int ix = indexOf(o);
		if(ix >= 0) {
			remove(ix);
			return true;
		} else {
			return false;
		}
	}
	public T remove(int ix) {
		Node<T> res = find(ix);
		res.remove(hat);
		return res.getValue();
	}
	public boolean removeAll(Collection<?> col) {
		boolean res = false;
		for(Object o : col) {
			res = res || remove(o);
		}
		return res;
	}
	public boolean retainAll(Collection<?> col) {
		boolean res = false;
		Node<T> tmp = hat.getHead();
		if(!col.contains(tmp.getValue())) {
			res = res || remove(tmp.getValue());
		}
		while(tmp.hasNext()) {
			tmp = tmp.getNext();
			if(!col.contains(tmp.getValue())) {
				res = res || remove(tmp.getValue());
			}
		}
		return res;
	}
	public T set(int ix, T o) {
		Node<T> tmp = find(ix);
		T res = tmp.getValue();
		tmp.setValue(o);
		return res;
	}
	public int size() {
		Node<T> tmp = hat.getHead();
		int res = 1;
		while(tmp.hasNext()) {
			res++;
			tmp = tmp.getNext();
		}
		return res;
	}
	public List<T> subList(int fromIndex, int toIndex) {
		List<T> res = new LinkedList<T>(get(fromIndex));
		for(int ix = fromIndex + 1; ix <= toIndex; ix++) {
			res.add(get(ix));
		}
		return res;
	}
	public Object[] toArray() {
		return toArray(new Object[size()]);
	}
	public <T2> T2[] toArray(T2[] a) {
		int ix = 0;
		Node<T> tmp = hat.getHead();
		a[ix] = (T2)tmp.getValue();
		ix++;
		while(tmp.hasNext()) {
			a[ix] = (T2)tmp.getValue();
			ix++;
			tmp = tmp.getNext();
		}
		return a;
	}
}
